/* Copyright 2022-2024 Romain Serra
 * Licensed to CS GROUP (CS) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * CS licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.orekit.tutorials.control.indirect;

import java.io.File;
import java.util.ArrayList;
import java.util.List;
import java.util.Locale;

import org.hipparchus.geometry.euclidean.threed.Vector3D;
import org.hipparchus.ode.ODEIntegrator;
import org.hipparchus.util.FastMath;
import org.orekit.attitudes.AttitudeProviderModifier;
import org.orekit.attitudes.FrameAlignedProvider;
import org.orekit.control.indirect.adjoint.CartesianAdjointDerivativesProvider;
import org.orekit.control.indirect.adjoint.CartesianAdjointEquationTerm;
import org.orekit.control.indirect.adjoint.CartesianAdjointKeplerianTerm;
import org.orekit.control.indirect.adjoint.cost.BoundedCartesianEnergy;
import org.orekit.control.indirect.adjoint.cost.CartesianCost;
import org.orekit.control.indirect.adjoint.cost.UnboundedCartesianEnergy;
import org.orekit.control.indirect.adjoint.cost.UnboundedCartesianEnergyNeglectingMass;
import org.orekit.control.indirect.shooting.NewtonFixedBoundaryCartesianSingleShooting;
import org.orekit.control.indirect.shooting.ShootingBoundaryOutput;
import org.orekit.control.indirect.shooting.boundary.CartesianBoundaryConditionChecker;
import org.orekit.control.indirect.shooting.boundary.FixedTimeBoundaryOrbits;
import org.orekit.control.indirect.shooting.boundary.NormBasedCartesianConditionChecker;
import org.orekit.control.indirect.shooting.propagation.CartesianAdjointDynamicsProvider;
import org.orekit.control.indirect.shooting.propagation.DormandPrince54IntegrationSettings;
import org.orekit.control.indirect.shooting.propagation.ShootingIntegrationSettings;
import org.orekit.control.indirect.shooting.propagation.ShootingPropagationSettings;
import org.orekit.data.DataContext;
import org.orekit.data.DataProvidersManager;
import org.orekit.data.DirectoryCrawler;
import org.orekit.forces.ForceModel;
import org.orekit.forces.gravity.NewtonianAttraction;
import org.orekit.frames.Frame;
import org.orekit.frames.FramesFactory;
import org.orekit.orbits.CartesianOrbit;
import org.orekit.orbits.OrbitType;
import org.orekit.propagation.AdditionalStateProvider;
import org.orekit.propagation.BoundedPropagator;
import org.orekit.propagation.EphemerisGenerator;
import org.orekit.propagation.SpacecraftState;
import org.orekit.propagation.conversion.ODEIntegratorBuilder;
import org.orekit.propagation.events.EventDetectionSettings;
import org.orekit.propagation.integration.AdditionalDerivativesProvider;
import org.orekit.propagation.numerical.NumericalPropagator;
import org.orekit.time.AbsoluteDate;
import org.orekit.time.DateTimeComponents;
import org.orekit.time.TimeScalesFactory;
import org.orekit.utils.Constants;
import org.orekit.utils.TimeStampedPVCoordinates;

/**
 * This class shows how to use the indirect (single) shooting method of Orekit.
 * The study case is inspired by the Global Trajectory Optimization Competition 12.
 * The goal is to find an arc between the Earth and an asteroid, with a thrust vector constrained in magnitude.
 * The problem is solved incrementally, by solving three instances, the final one being fully constrained.
 * @author Romain Serra
 * @since 12.2
 */
public class FixedBoundarySingleShooting {

    /** Sun GM. */
    private static final double MU = Constants.JPL_SSD_SUN_GM;

    /** Initial mass of the S/C (kg). */
    private static final double INITIAL_MASS = 2000.;

    /** ISP of the thrust (s). */
    private static final double SPECIFIC_IMPULSE = 4000.;

    /** Mass flow rate factor. */
    private static final double MASS_FLOW_RATE_FACTOR = 1. / (SPECIFIC_IMPULSE * Constants.G0_STANDARD_GRAVITY);

    /** Maximum thrust (N). */
    private static final double UPPER_BOUND_THRUST_MAGNITUDE = 0.6;
    /** Adjoint name. */
    private static final String ADJOINT_NAME = "adjoint";

    /** Thrust name. */
    private static final String THRUST_NAME = "thrust";

    // CHECKSTYLE: stop MultipleStringLiterals check
    /** Private empty constructor. */
    private FixedBoundarySingleShooting() { }

    /** Program entry point.
     * @param args program arguments
     */
    public static void main(final String[] args) {
        // configure Orekit
        configureData();

        // define boundary conditions
        final FixedTimeBoundaryOrbits boundaryOrbits = getOrbitalBoundary();
        // define tolerances to assert convergence
        final NormBasedCartesianConditionChecker conditionChecker = getConditionChecker();

        // start by neglecting mass changes and lifting the thrust magnitude constraint
        final ShootingBoundaryOutput shootingOutputUnboundedWithoutMass = solveWithUnboundedThrustAndNoMassFlow(boundaryOrbits,
                conditionChecker);
        if (!shootingOutputUnboundedWithoutMass.isConverged()) {
            System.out.println("Problem not solved.");
        }
        final double maximumThrustUnboundedWithoutMass = evaluateMaximumThrustMagnitude(shootingOutputUnboundedWithoutMass);
        System.out.println(maximumThrustUnboundedWithoutMass);

        // move on by putting back the mass in the problem
        final double[] initialAdjointUnboundedWithoutMass = shootingOutputUnboundedWithoutMass.getInitialState()
                .getAdditionalState(ADJOINT_NAME);
        final double[] guessUnbounded = buildUnboundedGuess(initialAdjointUnboundedWithoutMass);
        final EventDetectionSettings detectionSettings = getEventDetectionSettings();
        final ShootingBoundaryOutput shootingOutputUnbounded = solveWithUnboundedThrust(guessUnbounded, boundaryOrbits,
                conditionChecker, detectionSettings);
        if (!shootingOutputUnbounded.isConverged()) {
            System.out.println("Problem not solved.");
        }
        final double maximumThrustUnbounded = evaluateMaximumThrustMagnitude(shootingOutputUnbounded);
        System.out.println(maximumThrustUnbounded);

        // finally, add the thrust constraint
        final double[] initialAdjointUnbounded = shootingOutputUnbounded.getInitialState()
                .getAdditionalState(ADJOINT_NAME);
        final double[] guessBounded = initialAdjointUnbounded.clone();
        final ShootingBoundaryOutput shootingOutputBounded = solveWithBoundedThrust(guessBounded, boundaryOrbits,
                conditionChecker, detectionSettings);
        if (!shootingOutputBounded.isConverged()) {
            System.out.println("Problem not solved.");
        }
        final double maximumThrustBounded = evaluateMaximumThrustMagnitude(shootingOutputBounded);
        System.out.println(maximumThrustBounded);
    }

    private static void configureData() {
        final File home       = new File(System.getProperty("user.home"));
        final File orekitData = new File(home, "orekit-data");
        if (!orekitData.exists()) {
            System.err.format(Locale.US, "Failed to find %s folder%n",
                    orekitData.getAbsolutePath());
            System.err.format(Locale.US, "You need to download %s from %s, unzip it in %s and rename it 'orekit-data' for this tutorial to work%n",
                    "orekit-data-master.zip", "https://gitlab.orekit.org/orekit/orekit-data/-/archive/master/orekit-data-master.zip",
                    home.getAbsolutePath());
            System.exit(1);
        }

        final DataProvidersManager manager = DataContext.getDefault().getDataProvidersManager();
        manager.addProvider(new DirectoryCrawler(orekitData));
    }

    private static ShootingBoundaryOutput solveWithUnboundedThrustAndNoMassFlow(final FixedTimeBoundaryOrbits boundaryOrbits,
                                                                                final CartesianBoundaryConditionChecker conditionChecker) {
        final CartesianCost unboundedWithoutMass = new UnboundedCartesianEnergyNeglectingMass(ADJOINT_NAME);
        final NewtonFixedBoundaryCartesianSingleShooting shootingUnboundedWithoutMass = getShootingMethod(unboundedWithoutMass,
                boundaryOrbits, conditionChecker);
        // problem is simple enough to "guess" the initial adjoint as the zero vector (no control)
        final double[] guessUnboundedWithoutMass = new double[6];
        return shootingUnboundedWithoutMass.solve(INITIAL_MASS, guessUnboundedWithoutMass);
    }

    private static double[] buildUnboundedGuess(final double[] initialAdjointUnboundedWithoutMass) {
        // get the new guess (the control vector was the acceleration and now becomes the force)
        final double[] guessUnbounded = new double[7];
        for (int i = 0; i < initialAdjointUnboundedWithoutMass.length; i++) {
            guessUnbounded[i] = initialAdjointUnboundedWithoutMass[i] * (INITIAL_MASS * INITIAL_MASS);
        }
        guessUnbounded[6] = -1.;
        return guessUnbounded;
    }

    private static ShootingBoundaryOutput solveWithUnboundedThrust(final double[] guessUnbounded,
                                                                   final FixedTimeBoundaryOrbits boundaryOrbits,
                                                                   final CartesianBoundaryConditionChecker conditionChecker,
                                                                   final EventDetectionSettings detectionSettings) {
        final CartesianCost unbounded = new UnboundedCartesianEnergy(ADJOINT_NAME, MASS_FLOW_RATE_FACTOR, detectionSettings);
        final NewtonFixedBoundaryCartesianSingleShooting shootingUnbounded = getShootingMethod(unbounded, boundaryOrbits,
                conditionChecker);
        return shootingUnbounded.solve(INITIAL_MASS, guessUnbounded);
    }

    private static ShootingBoundaryOutput solveWithBoundedThrust(final double[] guessBounded,
                                                                 final FixedTimeBoundaryOrbits boundaryOrbits,
                                                                 final CartesianBoundaryConditionChecker conditionChecker,
                                                                 final EventDetectionSettings detectionSettings) {
        final CartesianCost bounded = new BoundedCartesianEnergy(ADJOINT_NAME, MASS_FLOW_RATE_FACTOR,
                UPPER_BOUND_THRUST_MAGNITUDE, detectionSettings);
        final NewtonFixedBoundaryCartesianSingleShooting shootingBounded = getShootingMethod(bounded, boundaryOrbits,
                conditionChecker);
        return shootingBounded.solve(INITIAL_MASS, guessBounded);
    }

    private static NewtonFixedBoundaryCartesianSingleShooting getShootingMethod(final CartesianCost cost,
                                                                                final FixedTimeBoundaryOrbits fixedTimeBoundaryOrbits,
                                                                                final CartesianBoundaryConditionChecker conditionChecker) {
        final NewtonFixedBoundaryCartesianSingleShooting shooting = new NewtonFixedBoundaryCartesianSingleShooting(getPropagationSettings(cost),
                fixedTimeBoundaryOrbits, conditionChecker);
        shooting.setScalePositionDefects(1e6);
        return shooting;
    }

    private static NormBasedCartesianConditionChecker getConditionChecker() {
        return new NormBasedCartesianConditionChecker(100, 1e5, 1e-1);
    }

    private static EventDetectionSettings getEventDetectionSettings() {
        return new EventDetectionSettings(1e4, 1e3, EventDetectionSettings.DEFAULT_MAX_ITER);
    }

    private static ShootingPropagationSettings getPropagationSettings(final CartesianCost cost) {
        // create external forces list with Keplerian dynamics only
        final List<ForceModel> forces = new ArrayList<>();
        forces.add(new NewtonianAttraction(MU));
        final CartesianAdjointEquationTerm equationTerm = new CartesianAdjointKeplerianTerm(MU);
        // add corresponding adjoint term to input cost and form derivatives provider
        final CartesianAdjointDynamicsProvider dynamicsProvider = new CartesianAdjointDynamicsProvider(cost,
                equationTerm);
        // add integration settings to form full propagation settings
        return new ShootingPropagationSettings(forces, dynamicsProvider, getIntegrationSettings());
    }

    private static ShootingIntegrationSettings getIntegrationSettings() {
        return new DormandPrince54IntegrationSettings(300, 3e5, 1e5, 1e-1);
    }

    private static FixedTimeBoundaryOrbits getOrbitalBoundary() {
        return new FixedTimeBoundaryOrbits(getInitialOrbit(), getTerminalOrbit());
    }

    private static CartesianOrbit getInitialOrbit() {
        // Earth
        final AbsoluteDate date = new AbsoluteDate(new DateTimeComponents(2035, 1, 1, 0, 0, 0),
                TimeScalesFactory.getUTC());
        final Vector3D position = new Vector3D(269630575634.1845, -317928797663.87445, -117503661424.1842);
        final Vector3D velocity = new Vector3D(12803.992418160833, 12346.009014593829, 2789.3378661767967);
        return new CartesianOrbit(new TimeStampedPVCoordinates(date, position, velocity), getFrame(), MU);
    }

    private static CartesianOrbit getTerminalOrbit() {
        // Asteroid 200
        final AbsoluteDate date = new AbsoluteDate(new DateTimeComponents(2038, 4, 20, 7, 48, 0),
                TimeScalesFactory.getUTC());
        final Vector3D position = new Vector3D(-254040098474.26975, 292309940514.6629, 61765199864.609174);
        final Vector3D velocity = new Vector3D(-15342.352873059252, -10427.635262141607, -7365.033285214819);
        return new CartesianOrbit(new TimeStampedPVCoordinates(date, position, velocity), getFrame(), MU);
    }

    private static Frame getFrame() {
        // trajectory is heliocentric yet there is no frame transformation, so let's keep Orekit "root" frame
        return FramesFactory.getGCRF();
    }

    private static double evaluateMaximumThrustMagnitude(final ShootingBoundaryOutput shootingBoundaryOutput) {
        final BoundedPropagator boundedPropagator = getIntegratedEphemeris(shootingBoundaryOutput);
        // add to integrated ephemeris the thrust computation
        final AdditionalDerivativesProvider adjointDerivativesProvider = shootingBoundaryOutput.getShootingPropagationSettings()
                .getAdjointDynamicsProvider().buildAdditionalDerivativesProvider();
        boundedPropagator.addAdditionalStateProvider(getThrustForceProvider((CartesianAdjointDerivativesProvider) adjointDerivativesProvider));
        // compute maximum magnitude on uniform time grid
        return evaluateMaximumThrustMagnitudeOnGrid(boundedPropagator);
    }

    private static double evaluateMaximumThrustMagnitudeOnGrid(final BoundedPropagator boundedPropagator) {
        final SpacecraftState initialState = boundedPropagator.propagate(boundedPropagator.getMinDate());
        final SpacecraftState terminalState = boundedPropagator.propagate(boundedPropagator.getMaxDate());
        final Vector3D terminalThrustForce = new Vector3D(terminalState.getAdditionalState(THRUST_NAME));
        double maximumMagnitude = terminalThrustForce.getNorm();
        for (double step = 0; step < boundedPropagator.getMaxDate().durationFrom(boundedPropagator.getMinDate()); step += 1e3) {
            final SpacecraftState state = boundedPropagator.propagate(initialState.getDate().shiftedBy(step));
            final Vector3D thrustForce = new Vector3D(state.getAdditionalState(THRUST_NAME));
            maximumMagnitude = FastMath.max(maximumMagnitude, thrustForce.getNorm());
        }
        return maximumMagnitude;
    }

    private static BoundedPropagator getIntegratedEphemeris(final ShootingBoundaryOutput shootingBoundaryOutput) {
        // set up propagation
        final ShootingPropagationSettings propagationSettings = shootingBoundaryOutput.getShootingPropagationSettings();
        final ODEIntegratorBuilder integratorBuilder = propagationSettings.getIntegrationSettings().getIntegratorBuilder();
        final SpacecraftState initialState = shootingBoundaryOutput.getInitialState();
        final OrbitType orbitType = OrbitType.CARTESIAN;
        final ODEIntegrator integrator = integratorBuilder.buildIntegrator(initialState.getOrbit(), orbitType);
        final NumericalPropagator propagator = new NumericalPropagator(integrator);
        propagator.setInitialState(initialState);
        propagator.setOrbitType(orbitType);
        final AdditionalDerivativesProvider adjointDerivativesProvider = propagationSettings.getAdjointDynamicsProvider()
                .buildAdditionalDerivativesProvider();
        propagator.addAdditionalDerivativesProvider(adjointDerivativesProvider);
        propagator.setAttitudeProvider(AttitudeProviderModifier.getFrozenAttitudeProvider(new FrameAlignedProvider(initialState.getFrame())));
        // propagate with dense output
        final EphemerisGenerator ephemerisGenerator = propagator.getEphemerisGenerator();
        propagator.propagate(shootingBoundaryOutput.getTerminalState().getDate());
        return ephemerisGenerator.getGeneratedEphemeris();
    }

    private static AdditionalStateProvider getThrustForceProvider(final CartesianAdjointDerivativesProvider derivativesProvider) {
        return new AdditionalStateProvider() {
            @Override
            public String getName() {
                return THRUST_NAME;
            }

            @Override
            public double[] getAdditionalState(final SpacecraftState state) {
                final double mass = state.getMass();
                final double[] adjoint = state.getAdditionalState(ADJOINT_NAME);
                final Vector3D thrustAcceleration = derivativesProvider.getCost().getThrustAccelerationVector(adjoint, mass);
                return thrustAcceleration.scalarMultiply(mass).toArray();
            }
        };
    }

}
