/* Copyright 2023 Luc Maisonobe
 * Licensed to CS GROUP (CS) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * CS licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.orekit.tutorials.propagation;

import java.io.File;
import java.io.IOException;
import java.io.PrintStream;
import java.net.URISyntaxException;
import java.net.URL;
import java.nio.charset.StandardCharsets;
import java.util.Locale;

import org.hipparchus.geometry.euclidean.threed.Vector3D;
import org.hipparchus.util.FastMath;
import org.orekit.bodies.GeodeticPoint;
import org.orekit.bodies.OneAxisEllipsoid;
import org.orekit.data.DataContext;
import org.orekit.data.DataProvidersManager;
import org.orekit.data.DirectoryCrawler;
import org.orekit.errors.OrekitException;
import org.orekit.orbits.Orbit;
import org.orekit.propagation.Propagator;
import org.orekit.propagation.SpacecraftState;
import org.orekit.propagation.analytical.KeplerianPropagator;
import org.orekit.propagation.sampling.OrekitFixedStepHandler;
import org.orekit.time.AbsoluteDate;
import org.orekit.tutorials.utils.OrekitBuilder;
import org.orekit.tutorials.yaml.TutorialBody;
import org.orekit.tutorials.yaml.TutorialOrbit;
import org.orekit.utils.Constants;

import com.fasterxml.jackson.annotation.JsonCreator;
import com.fasterxml.jackson.annotation.JsonProperty;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.dataformat.yaml.YAMLFactory;

/** Orekit tutorial for graphical display of ground tracks.
 * @author Luc Maisonobe
 */
//CHECKSTYLE: stop MultipleStringLiterals check
public class GroundTrack {

    /** Private constructor for utility class. */
    private GroundTrack() {
        // empty
    }

    /** Program entry point.
     * @param args program arguments
     */
    public static void main(final String[] args) {
        try {

            if (args.length != 1) {
                System.err.println("usage: java org.orekit.tutorials.propagation.GroundTrack ground-track.yaml");
                System.exit(1);
            }

            // configure Orekit
            final File home       = new File(System.getProperty("user.home"));
            final File orekitData = new File(home, "orekit-data");
            if (!orekitData.exists()) {
                System.err.format(Locale.US, "Failed to find %s folder%n",
                                  orekitData.getAbsolutePath());
                System.err.format(Locale.US, "You need to download %s from %s, unzip it in %s and rename it 'orekit-data' for this tutorial to work%n",
                                  "orekit-data-master.zip", "https://gitlab.orekit.org/orekit/orekit-data/-/archive/master/orekit-data-master.zip",
                                  home.getAbsolutePath());
                System.exit(1);
            }
            final DataProvidersManager manager = DataContext.getDefault().getDataProvidersManager();
            manager.addProvider(new DirectoryCrawler(orekitData));

            // input/out
            final URL url = GroundTrack.class.getResource("/" + args[0]);
            if (url == null) {
                System.err.println("resource " + args[0] + " not found");
                System.exit(1);
            }
            final File input  = new File(url.toURI().getPath());

            new GroundTrack().run(input);

        } catch (URISyntaxException | IOException | IllegalArgumentException | OrekitException e) {
            System.err.println(e.getLocalizedMessage());
            System.exit(1);
        }
    }

    /** run the program.
     * @param input input file
     * @throws IOException if input file cannot be read
     */
    private void run(final File input) throws IOException {

        // read input parameters
        final ObjectMapper mapper = new ObjectMapper(new YAMLFactory());
        mapper.findAndRegisterModules();
        final TutorialGroundTrack inputData = mapper.readValue(input, TutorialGroundTrack.class);

        // spaceflight dynamics objects
        final Orbit            orbit = OrekitBuilder.createOrbit(inputData.orbit, Constants.EIGEN5C_EARTH_MU);
        final OneAxisEllipsoid earth = OrekitBuilder.createBody(inputData.body);

        final ProcessBuilder pb = new ProcessBuilder("gnuplot").
                                  redirectOutput(ProcessBuilder.Redirect.INHERIT).
                                  redirectError(ProcessBuilder.Redirect.INHERIT);
        pb.environment().remove("XDG_SESSION_TYPE");
        final Process gnuplot = pb.start();
        try (PrintStream out = new PrintStream(gnuplot.getOutputStream(), false, StandardCharsets.UTF_8.name())) {
            if (inputData.outputName == null) {
                out.format(Locale.US, "set terminal qt size %d, %d title 'complex plotter'%n", inputData.width, inputData.height);
            } else {
                out.format(Locale.US, "set terminal pngcairo size %d, %d%n", inputData.width, inputData.height);
                out.format(Locale.US, "set output '%s'%n", new File(inputData.outputName).getAbsolutePath());
            }
            if (inputData.generate3D) {
                out.format(Locale.US, "unset tics%n");
                out.format(Locale.US, "unset colorbox%n");
                out.format(Locale.US, "set margins 0, 0, 0, 0%n");
                out.format(Locale.US, "set border 0%n");
                out.format(Locale.US, "set mapping spherical%n");
                out.format(Locale.US, "set angles degrees%n");
                out.format(Locale.US, "set view %f, %f%n",
                           90 - inputData.observerLatitude,
                           inputData.observerLongitude < -90 ? (450 + inputData.observerLongitude) :
                           (inputData.observerLongitude > 270 ? inputData.observerLongitude - 270 : 90 + inputData.observerLongitude));
            } else {
                out.format(Locale.US, "set size ratio -1%n");
                out.format(Locale.US, "set grid front lc -1 dt 3 lw 1 lt 1%n");
                out.format(Locale.US, "set xtics 45%n");
                out.format(Locale.US, "set ytics 30%n");
                out.format(Locale.US, "set xrange [%f : %f]%n", -180.0, 180.0);
                out.format(Locale.US, "set xlabel 'longitude (°)'%n");
                out.format(Locale.US, "set yrange [%f : %f]%n",  -90.0,  90.0);
                out.format(Locale.US, "set ylabel 'latitude (°)'%n");
            }
            out.format(Locale.US, "set key off%n");
            out.format(Locale.US, "set title '%s'%n", inputData.title);
            out.format(Locale.US, "$data <<EOD%n");

            final OrekitFixedStepHandler handler = inputData.generate3D ?
                                                   new ThreeDHandler(earth,
                                                                     new GeodeticPoint(FastMath.toRadians(inputData.observerLatitude),
                                                                                       FastMath.toRadians(inputData.observerLongitude),
                                                                                       0.0),
                                                                     out) :
                                                   new TwoDHandler(earth, out);
            final Propagator propagator = new KeplerianPropagator(orbit);
            propagator.getMultiplexer().add(inputData.outputStep, handler);
            propagator.propagate(orbit.getDate().shiftedBy(inputData.nbOrbits * orbit.getKeplerianPeriod()));

            out.format(Locale.US, "EOD%n");
            out.format(Locale.US, "set datafile separator ','%n");
            out.format(Locale.US, "rgb(r,g,b) = 65536 * int(r) + 256 * int(g) + int(b)%n");
            if (inputData.generate3D) {
                out.format(Locale.US, "splot '%s' using 1:2:(1):(rgb($3,$4,$5)) with pm3d lc rgb variable,\\%n",
                           new File(input.getParent(), inputData.naturalEarthRaster));
                out.format(Locale.US, "      $data using 1:2 with lines lw 1.5 lc rgb 'goldenrod' notitle%n");
            } else {
                out.format(Locale.US, "plot '%s' using 1:2:(rgb($3,$4,$5)) with rgbimage,\\%n",
                           new File(input.getParent(), inputData.naturalEarthRaster));
                out.format(Locale.US, "      $data using 1:2 with lines lw 1.5 lc rgb 'goldenrod' notitle%n");
            }
            if (inputData.outputName == null) {
                out.format(Locale.US, "pause mouse close%n");
            } else {
                System.out.format(Locale.US, "output written to %s%n",
                                  new File(inputData.outputName).getAbsolutePath());
            }
        }

    }

    static class TwoDHandler implements OrekitFixedStepHandler {

        /** Earth ellipsoid. */
        private final OneAxisEllipsoid earth;

        /** Output stream. */
        private final PrintStream out;

        /** Previous point on ellipsoid. */
        private GeodeticPoint previous;

        TwoDHandler(final OneAxisEllipsoid earth, final PrintStream out) {
            this.earth = earth;
            this.out   = out;
        }
        public void init(final SpacecraftState s0, final AbsoluteDate t) {
            previous = null;
        }
        public void handleStep(final SpacecraftState state) {
            final GeodeticPoint gp = earth.transform(state.getPosition(earth.getBodyFrame()),
                                                     earth.getBodyFrame(),
                                                     state.getDate());
            if (previous != null && FastMath.abs(gp.getLongitude() - previous.getLongitude()) > FastMath.PI) {
                // add two empty lines to split line and avoid long strokes when passing antimeridian
                out.format(Locale.US, "%n%n");
            }
            out.format(Locale.US, "%.3f,%.3f%n",
                       FastMath.toDegrees(gp.getLongitude()),
                       FastMath.toDegrees(gp.getLatitude()));
            previous = gp;
        }
    }

    static class ThreeDHandler implements OrekitFixedStepHandler {

        /** Earth ellipsoid. */
        private final OneAxisEllipsoid earth;

        /** Output stream. */
        private final PrintStream out;

        /** Observer direction. */
        private final Vector3D observer;

        /** Plotting flag. */
        private boolean plotting;

        ThreeDHandler(final OneAxisEllipsoid earth, final GeodeticPoint observer, final PrintStream out) {
            this.earth    = earth;
            this.observer = earth.transform(observer);
            this.out      = out;
            this.plotting = true;
        }
        public void handleStep(final SpacecraftState state) {
            final Vector3D point = state.getPosition(earth.getBodyFrame());
            if (Vector3D.dotProduct(point, observer) <= 0 && plotting) {
                out.format(Locale.US, "%n%n");
                plotting = false;
            } else {
                final GeodeticPoint gp = earth.transform(point, earth.getBodyFrame(), state.getDate());
                out.format(Locale.US, "%.3f,%.3f%n",
                           FastMath.toDegrees(gp.getLongitude()),
                           FastMath.toDegrees(gp.getLatitude()));
                plotting = true;
            }
        }
    }

    /**
     * Initial data to initialize a ground track tutorial.
     * <p>
     * Data are read from a YAML file.
     * </p>
     * @author Luc Maisonobe
     */
    static class TutorialGroundTrack {

        /** Data used to initialize the central body. */
        private final TutorialBody body;

        /** Data used to initialize the initial orbit. */
        private final TutorialOrbit orbit;

        /** Name of the output file, display will be on screen if empty. */
        private final String outputName;

        /** Output step size. */
        private final double outputStep;

        /** Number of orbits to display (not necessarily an integer). */
        private final double nbOrbits;

        /** Plot title. */
        private final String title;

        /** Width of the plot. */
        private final int width;

        /** Height of the plot. */
        private final int height;

        /** Flag for the 3D generation. */
        private final boolean generate3D;

        /** Observer longitude. */
        private final double observerLongitude;

        /** Observer latitude. */
        private final double observerLatitude;

        /** Name of natural Earth raster file. */
        private final String naturalEarthRaster;

        @JsonCreator
        TutorialGroundTrack(@JsonProperty("body")               final TutorialBody body,
                            @JsonProperty("orbit")              final TutorialOrbit orbit,
                            @JsonProperty("outputName")         final String outputName,
                            @JsonProperty("outputStep")         final double outputStep,
                            @JsonProperty("nbOrbits")           final double nbOrbits,
                            @JsonProperty("title")              final String title,
                            @JsonProperty("width")              final int width,
                            @JsonProperty("height")             final int height,
                            @JsonProperty("generate3D")         final boolean generate3D,
                            @JsonProperty("observerLongitude")  final double observerLongitude,
                            @JsonProperty("observerLatitude")   final double observerLatitude,
                            @JsonProperty("naturalEarthRaster") final String naturalEarthRaster) {
            this.body               = body;
            this.orbit              = orbit;
            this.outputName         = outputName;
            this.outputStep         = outputStep;
            this.nbOrbits           = nbOrbits;
            this.title              = title;
            this.width              = width;
            this.height             = height;
            this.generate3D         = generate3D;
            this.observerLongitude  = observerLongitude;
            this.observerLatitude   = observerLatitude;
            this.naturalEarthRaster = naturalEarthRaster;
        }

    }

}
//CHECKSTYLE: resume MultipleStringLiterals check

