/* Copyright 2002-2023 CS GROUP
 * Licensed to CS GROUP (CS) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * CS licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.orekit.tutorials.utils;

import org.hipparchus.geometry.euclidean.threed.Vector3D;
import org.hipparchus.util.FastMath;
import org.orekit.bodies.OneAxisEllipsoid;
import org.orekit.frames.Frame;
import org.orekit.frames.FramesFactory;
import org.orekit.orbits.CartesianOrbit;
import org.orekit.orbits.CircularOrbit;
import org.orekit.orbits.EquinoctialOrbit;
import org.orekit.orbits.KeplerianOrbit;
import org.orekit.orbits.Orbit;
import org.orekit.orbits.PositionAngleType;
import org.orekit.propagation.SpacecraftState;
import org.orekit.propagation.analytical.tle.TLE;
import org.orekit.propagation.analytical.tle.TLEPropagator;
import org.orekit.time.AbsoluteDate;
import org.orekit.time.TimeScalesFactory;
import org.orekit.tutorials.yaml.TutorialBody;
import org.orekit.tutorials.yaml.TutorialOrbit;
import org.orekit.tutorials.yaml.TutorialOrbitType;
import org.orekit.tutorials.yaml.TutorialOrbitType.TutorialCartesianOrbit;
import org.orekit.tutorials.yaml.TutorialOrbitType.TutorialCircularOrbit;
import org.orekit.tutorials.yaml.TutorialOrbitType.TutorialEquinoctialOrbit;
import org.orekit.tutorials.yaml.TutorialOrbitType.TutorialKeplerianOrbit;
import org.orekit.tutorials.yaml.TutorialOrbitType.TutorialTLE;
import org.orekit.utils.Constants;
import org.orekit.utils.IERSConventions;
import org.orekit.utils.PVCoordinates;

/**
 * Utility class for building Orekit objects from parsed YAML files.
 *
 * @author Pascal Parraud
 * @author Bryan Cazabonne
 * @author Luc Maisonobe
 *
 */
public class OrekitBuilder {

    /** Null double values are equal to 0.0 in YAML file. */
    private static final double NULL_DOUBLE = 0.0;

    /** Private constructor for this utility class. */
    private OrekitBuilder() {
       // Do nothing
    }

    /** Create a central body.
     * @param body central body block from a parsed file
     * @return central body
     */
    public static OneAxisEllipsoid createBody(final TutorialBody body) {

        final Frame bodyFrame;
        if (body.getFrameName() != null) {
            bodyFrame = body.getEarthFrame();
        } else {
            bodyFrame = FramesFactory.getITRF(IERSConventions.IERS_2010, true);
        }

        final double equatorialRadius;
        if (body.getEquatorialRadius() != NULL_DOUBLE) {
            equatorialRadius = body.getEquatorialRadius();
        } else {
            equatorialRadius = Constants.WGS84_EARTH_EQUATORIAL_RADIUS;
        }

        final double flattening;
        if (body.getInverseFlattening() != NULL_DOUBLE) {
            flattening = 1.0 / body.getInverseFlattening();
        } else {
            flattening = Constants.WGS84_EARTH_FLATTENING;
        }

        return new OneAxisEllipsoid(equatorialRadius, flattening, bodyFrame);

    }

    /** Create an orbit from input parameters.
     * @param orbit orbit block from a parsed file
     * @param mu central attraction coefficient
     * @return orbit
     */
    public static Orbit createOrbit(final TutorialOrbit orbit, final double mu) {

        final TutorialOrbitType data  = orbit.getOrbitType();

        // inertial frame
        final Frame frame = orbit.getInertialFrame() == null ? FramesFactory.getEME2000() : orbit.getInertialFrame();

        // orbit definition
        if (data.getKeplerian() != null) {
            final TutorialKeplerianOrbit keplerian = data.getKeplerian();
            return new KeplerianOrbit(keplerian.getA(),
                                      keplerian.getE(),
                                      FastMath.toRadians(keplerian.getI()),
                                      FastMath.toRadians(keplerian.getPa()),
                                      FastMath.toRadians(keplerian.getRaan()),
                                      FastMath.toRadians(keplerian.getV()),
                                      PositionAngleType.valueOf(keplerian.getPositionAngle()),
                                      frame,
                                      new AbsoluteDate(orbit.getDate(),
                                                       TimeScalesFactory.getUTC()),
                                      mu);
        } else if (data.getEquinoctial() != null) {
            final TutorialEquinoctialOrbit equinoctial = data.getEquinoctial();
            return new EquinoctialOrbit(equinoctial.getA(),
                                        equinoctial.getEx(),
                                        equinoctial.getEy(),
                                        equinoctial.getHx(),
                                        equinoctial.getHy(),
                                        FastMath.toRadians(equinoctial.getLv()),
                                        PositionAngleType.valueOf(equinoctial.getPositionAngle()),
                                        frame,
                                        new AbsoluteDate(orbit.getDate(),
                                                         TimeScalesFactory.getUTC()),
                                        mu);
        } else if (data.getCircular() != null) {
            final TutorialCircularOrbit circular = data.getCircular();
            return new CircularOrbit(circular.getA(),
                                     circular.getEx(),
                                     circular.getEy(),
                                     FastMath.toRadians(circular.getI()),
                                     FastMath.toRadians(circular.getRaan()),
                                     FastMath.toRadians(circular.getAlphaV()),
                                     PositionAngleType.valueOf(circular.getPositionAngle()),
                                     frame,
                                     new AbsoluteDate(orbit.getDate(),
                                                      TimeScalesFactory.getUTC()),
                                     mu);
        } else if (data.getTle() != null) {
            final TutorialTLE tleData = data.getTle();
            final String line1 = tleData.getLine1();
            final String line2 = tleData.getLine2();
            final TLE tle = new TLE(line1, line2);

            final TLEPropagator propagator = TLEPropagator.selectExtrapolator(tle);

            final AbsoluteDate initDate = tle.getDate();
            final SpacecraftState initialState = propagator.getInitialState();


            //Transformation from TEME to frame.
            return new CartesianOrbit(initialState.getPVCoordinates(frame),
                                      frame,
                                      initDate,
                                      mu);


        } else {
            final TutorialCartesianOrbit cartesian = data.getCartesian();
            final double[] pos = {cartesian.getX(), cartesian.getY(), cartesian.getZ()};
            final double[] vel = {cartesian.getVx(), cartesian.getVy(), cartesian.getVz()};

            return new CartesianOrbit(new PVCoordinates(new Vector3D(pos), new Vector3D(vel)),
                                      frame,
                                      new AbsoluteDate(orbit.getDate(),
                                                       TimeScalesFactory.getUTC()),
                                      mu);
        }
    }

}
